import re
import inspect
import traceback
import linecache
try:
    from collections.abc import MutableMapping  # noqa
except ImportError:
    from collections import MutableMapping  # noqa


class TestResult:
    def __init__(self, test_status=None, test_name=None, filename=None, desired_caps=None, appium_log=None,
                 screenshot=None, page_source=None, error=None, code_context=None, elapsed_time_ms=None, flutter_render_tree=None):
        self.status = test_status
        self.name = test_name
        self.filename = filename
        self.desired_caps = desired_caps
        self.log = appium_log
        self.screenshot = screenshot
        self.page_source = page_source
        self.error = error
        self.code_context = code_context
        self.elapsed_time_ms = elapsed_time_ms
        self.flutter_render_tree = flutter_render_tree


def format_error(error_type=None, error_msg=None, line_number=None, invoked_func=None, code_executed=None):
    return {
        'error_type': error_type,
        'error_msg': error_msg,
        'line_number': line_number,
        'invoked_func': invoked_func,
        'code_executed': code_executed
    }


def format_appium_log(log_list):
    output = []
    for event in log_list:
        # ANSI escape sequences
        # https://bit.ly/3rK88pe
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')

        # Split at first event type occurence
        # in square brackets, e.g. [HTTP]
        # https://bit.ly/3fItHEi
        split_log_msg = re.split(r'\[|\]', ansi_escape.sub(
            '', event['message']), maxsplit=2)

        formatted_event = {
            'timestamp': event.get('timestamp'),
            'level': event.get('level'),
            'type': split_log_msg[1],
            'message': split_log_msg[2].strip()
        }

        output.append(formatted_event)
    return output


def flatten_desired_capabilities(desired_caps, parent_key='', sep='_'):
    items = []
    for key, value in desired_caps.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, MutableMapping):
            items.extend(flatten_desired_capabilities(
                value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)


def get_automation_type(desired_caps):
    if desired_caps.get('automationName'):
        return 'Appium'
    elif desired_caps.get('browserName'):
        return 'Selenium'
    else:
        return None


def get_platform_name(desired_caps):
    if desired_caps.get('browserName'):
        return desired_caps.get('browserName').title()
    elif desired_caps.get('platformName'):
        mobilePlatformName = desired_caps.get('platformName').lower()
        if mobilePlatformName == "ios":
            return "iOS"
        else:
            return mobilePlatformName.title()
    else:
        return None


def get_platform_version(desired_caps):
    if desired_caps.get('browserVersion'):
        return desired_caps.get('browserVersion')
    elif desired_caps.get('platformVersion'):
        return desired_caps.get('platformVersion')
    else:
        return None


def validate_test_name(custom_test_name):
    if not custom_test_name or not type(custom_test_name) is str or len(custom_test_name) == 0:
        return None

    test_name = custom_test_name.strip()

    if len(test_name) >= 43:
        return test_name[:40] + "..."
    else:
        return test_name


def calculate_elapsed_time_ms(start_time, end_time):
    return int((end_time - start_time) * 1000)


def get_testcase_steps(testcase, error_line_number=0):
    testcase_source_raw = inspect.getsourcelines(testcase)
    testcase_source_clean = [step.strip() for step in testcase_source_raw[0]]
    line_number = testcase_source_raw[1]

    # Discard the initial 2 lines as they are the decorator and the testcase function name
    testcase_source_clean = testcase_source_clean[2:]
    line_number += 2

    output = []
    for step in testcase_source_clean:
        output.append(
            {
                'line_number': line_number,
                'line_code': step
            }
        )
        line_number += 1

    if error_line_number > 0:
        for index, value in enumerate(output):
            if value['line_number'] == error_line_number:
                # get previous 9 lines plus the line where the error occurred
                # ensure that the start range value is never below zero
                # get the next 9 lines after the error occured
                # ensure that the end range value never exceeds the len of the list
                result = output[max(index - 9, 0): min(index + 10, len(output))]
                return result
    else:
        return output


def print_modified_exception_traceback(exc_type, exc_value, exc_traceback, tauk_package_filename):
    # Build a new stack with the traceback objects,
    # except the one traceback object that references the Tauk package
    new_traceback_stack = []
    while exc_traceback is not None:
        # Skip frames that contain the Tauk package file ('tauk_appium.py')
        # in the filename
        if tauk_package_filename in exc_traceback.tb_frame.f_code.co_filename:
            exc_traceback = exc_traceback.tb_next
            continue

        new_traceback_stack.append(exc_traceback)
        exc_traceback = exc_traceback.tb_next

    # Assign tb_next in reverse to avoid circular references
    tb_next = None
    for tb in reversed(new_traceback_stack):
        tb.tb_next = tb_next
        tb_next = tb

    print(''.join(traceback.format_exception(exc_type, exc_value, tb_next)))


class FuncTracer:
    '''Usage:
    1. Create an instance
    func_tracer = FuncTracer(caller_filename)

    2. Add the function name you want to trace to the list
    func_tracer.TRACE_INTO.append(func.__name__)

    3. Pass the instance into sys.settrace
    sys.settrace(func_tracer.trace_calls)
    '''

    def __init__(self, filename):
        self.filename = filename
        self.TRACE_INTO = []
        self.EXECUTED_LINE_NUMBERS = set([])
        self.EXECUTED_STEPS = []

    def trace_lines(self, frame, event, arg):
        if event != 'line':
            return

        code_obj = frame.f_code
        func_name = code_obj.co_name
        line_no = frame.f_lineno
        filename = code_obj.co_filename

        line = linecache.getline(filename, line_no)

        if not 'print' in line.strip() and not line_no in self.EXECUTED_LINE_NUMBERS:
            self.EXECUTED_STEPS.append((line_no, line.strip()))
            self.EXECUTED_LINE_NUMBERS.add(line_no)

    def trace_calls(self, frame, event, arg):
        if event != 'call':
            return
        code_obj = frame.f_code
        func_name = code_obj.co_name
        if func_name == 'write':
            # ignore write() calls from print statements
            return

        if func_name == 'click':
            print(f'CLICK WAS CALLED ON LINE: {frame.f_back.f_lineno}')

        if func_name == 'send_keys':
            print("SEND KEYS WAS CALLED")

        if func_name == 'wait':
            print('WAIT WAS CALLED')

        line_no = frame.f_lineno
        filename = code_obj.co_filename

        # print(f"Call to {func_name} on line {line_no} of {filename}")
        if func_name in self.TRACE_INTO:
            # Trace into this function
            print(f"Call to {func_name} on line {line_no} of {filename}")
            return self.trace_lines
        return

    def trace_calls_and_returns(self, frame, event, arg):
        code_obj = frame.f_code
        func_name = code_obj.co_name
        if func_name == 'write':
            # Ignore write() calls from print statements
            return
        line_no = frame.f_lineno
        filename = code_obj.co_filename
        if event == 'call':
            print(f'Call to {func_name} on line {line_no} of {filename}')
            return self.trace_calls_and_returns
        elif event == 'return':
            print(f'{func_name} => {arg}')
        return



"""Helper package to facilitate reporting for webdriver-based tests on Tauk"""

from tauk.tauk_appium import Tauk

__project__ = "tauk"
__version__ = "1.0.9"
__author__ = "Nathan Krishnan"
__url__ = "https://github.com/thetauk/tauk-webdriver-python"
__platforms__ = "ALL"
__classifiers__ = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]
__requires__ = ["requests", "Appium-Python-Client", "selenium"]

__extra_requires__ = {
}



import sys
import time
import ntpath
import logging
import inspect
import traceback
from functools import partial
from tauk.enums import TestStatusType
from tauk.utils import (TestResult, format_appium_log, format_error, get_testcase_steps,
                        flatten_desired_capabilities, get_automation_type, calculate_elapsed_time_ms,
                        get_platform_name, get_platform_version, print_modified_exception_traceback,
                        validate_test_name
                        )
import requests
from requests.exceptions import HTTPError, ConnectionError, Timeout, RequestException


class Tauk:

    logging.basicConfig(
        filename='.tauk-appium-sdk.log',
        level=logging.WARNING,
        filemode='w',
        format='%(asctime)s - %(levelname)s - %(message)s'
    )

    @classmethod
    def initialize(cls, **kwargs):
        """Initializes the Tauk SDK
        :param api_token: Your API Token from the Tauk Web Application
        :param project_id: Your Project ID from the Tauk Web Application
        :param driver: The Appium driver object
        """

        required_args = ['api_token', 'project_id']

        for arg in required_args:
            if arg not in kwargs:
                raise TypeError(f'Argument {arg} is required')

        cls._api_token = kwargs.get('api_token')
        cls._project_id = kwargs.get('project_id')
        cls._driver = kwargs.get('driver')
        cls._excluded = True if kwargs.get('excluded') else False
        cls._test_results = []

    @classmethod
    def _get_page_source(cls):
        if cls._driver:
            try:
                if 'FLUTTER' in cls._driver.contexts:
                    cls._driver.switch_to.context('NATIVE_APP')

                raw_page_source = cls._driver.page_source
            except:
                logging.error(
                    "An issue occurred while requesting the page source.")
                logging.error(traceback.format_exc())
                return None
            return raw_page_source
        else:
            return None

    @classmethod
    def _get_screenshot(cls):
        if cls._driver:
            try:
                screenshot = cls._driver.get_screenshot_as_base64()
            except:
                logging.error(
                    "An issue occurred while trying to take a screenshot.")
                logging.error(traceback.format_exc())
                return None
            return screenshot
        else:
            return None

    @classmethod
    def _get_log(cls):
        # Get last 50 log entries
        # minus the 5 log entries for issuing get_log()
        slice_range = slice(-55, -5)

        if cls._driver:
            try:
                log = cls._driver.get_log('server')[slice_range]
            except:
                logging.error(
                    "An issue occurred while requesting the Appium server logs.")
                logging.error(traceback.format_exc())
                return None
            return log
        else:
            return None

    @classmethod
    def _get_desired_capabilities(cls):
        if cls._driver:
            try:
                # Appium
                dc = cls._driver.capabilities['desired']
            except:
                try:
                    # Selenium
                    dc = flatten_desired_capabilities(
                        cls._driver.capabilities)
                except:
                    logging.error(
                        "An issue occurred while trying to retrieve the desired capabilities.")
                    logging.error(traceback.format_exc())
                    return None
            return dc
        else:
            return None

    # Flutter Specific
    @classmethod
    def _get_flutter_render_tree(cls):
        if cls._driver and 'appium' in cls._driver.__class__.__module__ and 'FLUTTER' in cls._driver.contexts:
            try:
                cls._driver.switch_to.context('FLUTTER')
                render_tree_source = cls._driver.execute_script(
                    "flutter: getRenderTree")
            except:
                logging.error(
                    "An issue occurred while trying to retrieve the Flutter render tree.")
                logging.error(traceback.format_exc())
                return None
            return render_tree_source
        else:
            return None

    @classmethod
    def observe(cls, func=None, *, custom_test_name=None):

        if func is None:
            return partial(cls.observe, custom_test_name=custom_test_name)

        all_frames = inspect.stack()
        caller_frame = None
        caller_filename = None
        for frame_info in all_frames:
            if func.__name__ in frame_info.frame.f_code.co_names:
                caller_frame = frame_info
                caller_filename = caller_frame.filename
                break

        def invoke_test_case(*args, **kwargs):

            user_provided_test_name = custom_test_name

            try:
                start_time = time.perf_counter()
                result = func(*args, **kwargs)
            except:
                failure_end_time = time.perf_counter()

                exc_type, exc_value, exc_traceback = sys.exc_info()
                stack_summary_list = traceback.extract_tb(exc_traceback)
                filename, line_number, invoked_func, code_executed = None, None, None, None
                for stack_trace in stack_summary_list:
                    if stack_trace.filename == caller_filename:
                        filename, line_number, invoked_func, code_executed = stack_trace
                        break

                testcase_steps = get_testcase_steps(
                    testcase=func,
                    error_line_number=line_number
                )

                test_result = TestResult(
                    test_status=TestStatusType.excluded.name if cls._excluded else TestStatusType.failed.name,
                    test_name=validate_test_name(user_provided_test_name) or func.__name__,
                    filename=ntpath.basename(caller_filename),
                    desired_caps=cls._get_desired_capabilities(),
                    appium_log=cls._get_log(),
                    page_source=cls._get_page_source(),
                    error=format_error(
                        error_type=str(exc_value.__class__.__name__),
                        error_msg=str(exc_value),
                        line_number=int(line_number),
                        invoked_func=str(invoked_func),
                        code_executed=str(code_executed)
                    ),
                    code_context=testcase_steps,
                    elapsed_time_ms=calculate_elapsed_time_ms(
                        start_time, failure_end_time),
                    flutter_render_tree=cls._get_flutter_render_tree()
                )
                test_result.screenshot = cls._get_screenshot()
                cls._test_results.append(test_result)

                # Suppress the traceback information
                # (only the exception type and value will be printed)
                # and print our custom traceback information
                print_modified_exception_traceback(
                    exc_type, exc_value, exc_traceback, tauk_package_filename=__file__)
                sys.tracebacklimit = 0
                raise
            else:
                success_end_time = time.perf_counter()

                test_result = TestResult(
                    test_status=TestStatusType.excluded.name if cls._excluded else TestStatusType.passed.name,
                    test_name=validate_test_name(user_provided_test_name) or func.__name__,
                    filename=ntpath.basename(caller_filename),
                    desired_caps=cls._get_desired_capabilities(),
                    appium_log=cls._get_log(),
                    page_source=cls._get_page_source(),
                    error=None,
                    code_context=None,
                    elapsed_time_ms=calculate_elapsed_time_ms(
                        start_time, success_end_time),
                    flutter_render_tree=cls._get_flutter_render_tree()
                )
                test_result.screenshot = cls._get_screenshot()
                cls._test_results.append(test_result)

                return result

        return invoke_test_case

    @classmethod
    def upload(cls, custom_session_upload_url=None):
        headers = {
            'api_token': cls._api_token,
            'project_id': cls._project_id
        }

        if len(cls._test_results) > 0:
            for test_result in cls._test_results:

                payload = {
                    'test_status': test_result.status,
                    'test_name': test_result.name,
                    'test_filename': test_result.filename,
                    'tags': test_result.desired_caps,
                    'log': format_appium_log(test_result.log) if test_result.log is not None else None,
                    'screenshot': test_result.screenshot,
                    'view': test_result.page_source,
                    'error': test_result.error,
                    'code_context': test_result.code_context,
                    'automation_type': get_automation_type(test_result.desired_caps),
                    'language': 'Python',
                    'platform': get_platform_name(test_result.desired_caps),
                    'platform_version': get_platform_version(cls._driver.capabilities),
                    'elapsed_time_ms': test_result.elapsed_time_ms,
                    'flutter_render_tree': test_result.flutter_render_tree
                }

                try:
                    response = requests.request(
                        "POST",
                        custom_session_upload_url if custom_session_upload_url else 'https://www.tauk.com/api/v1/session/upload',
                        headers=headers,
                        json=payload,
                        timeout=(15, 15)
                    )
                    response.raise_for_status()

                except HTTPError as http_status_error:
                    logging.error("An HTTP status code error occurred.")
                    logging.error(http_status_error)
                    logging.error(http_status_error.response.text)
                    try:
                        print(
                            f"ERROR: {http_status_error.response.json().get('error')}")
                    except:
                        print(f"ERROR: {http_status_error.response.text}")
                except ConnectionError as connection_error:
                    logging.error("An error connecting to the API occurred.")
                    logging.error(connection_error)
                    logging.error(connection_error.text)
                    print("ERROR: Unable to connect to the Tauk Platform.")
                except Timeout as timeout_error:
                    logging.error("A timeout error occurred.")
                    logging.error(timeout_error)
                    logging.error(timeout_error.text)
                    print("ERROR: The upload request timed out.")
                except RequestException as request_error:
                    logging.error(
                        "An error occurred trying to make an upload request.")
                    logging.error(request_error)
                    logging.error(request_error.text)
                    print(f"ERROR: Unable to perform an upload request.")
                else:
                    return response.json()



"""Setup script for the Tauk package"""

from setuptools import setup, find_packages
# Read the contents of the README.md file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


def main():
    """Executes setup when this script is the top-level"""
    import tauk as app

    setup(
        name=app.__project__,
        version=app.__version__,
        description=app.__doc__,
        long_description=long_description,
        long_description_content_type='text/markdown',
        classifiers=app.__classifiers__,
        author=app.__author__,
        url=app.__url__,
        license=[
            c.rsplit('::', 1)[1].strip()
            for c in app.__classifiers__
            if c.startswith('License ::')
        ][0],
        packages=find_packages(),
        include_package_data=True,
        platforms=app.__platforms__,
        install_requires=app.__requires__,
        extras_require=app.__extra_requires__,
    )


if __name__ == '__main__':
    main()
